from typing import Optional

from b_cfn_lambda_layer.package_version import PackageVersion


class Dependency:
    def __init__(self, name: str, version: Optional[PackageVersion] = None):
        self.__name = name
        self.__version = version or PackageVersion.latest()

    def build_string(self) -> str:
        """
        Creates string representation of the dependency that can be used by PIP tool.

        For example, if you initiate this class:
        >>> dep = Dependency('jwt', PackageVersion.from_string_version('123'))
        Then when you execute this method, you will get:
        >>> dep.build_string() == 'jwt==123'

        :return: String representation.
        """
        if self.__version.version_type == PackageVersion.VersionType.LATEST:
            return self.__name

        if self.__version.version_type == PackageVersion.VersionType.SPECIFIC:
            return f'{self.__name}=={self.__version.version_string}'

        if self.__version.version_type == PackageVersion.VersionType.NONE:
            return ''

        else:
            raise ValueError('Unsupported enum value.')



import os
import shutil
from typing import Optional, List

from aws_cdk.aws_lambda import Code

from b_cfn_lambda_layer import root
from b_cfn_lambda_layer.dependency import Dependency
from b_cfn_lambda_layer.pip_install import PipInstall
from b_cfn_lambda_layer.tmp import docker_build_root


class LambdaLayerCode:
    DEFAULT_DOCKER_IMAGE = 'python:3.9'

    def __init__(
            self,
            source_path: str,
            additional_pip_install_args: Optional[str] = None,
            dependencies: Optional[List[Dependency]] = None,
            docker_image: Optional[str] = None
    ) -> None:
        """
        Constructor.

        :param source_path: Path to source-code to be bundled.
        :param additional_pip_install_args:
        :param dependencies:
        :param docker_image:
        """
        self.additional_pip_install_args = additional_pip_install_args
        self.dependencies = dependencies
        self.source_path = source_path
        self.source_path_dir_name = os.path.basename(self.source_path)
        self.docker_image = docker_image or self.DEFAULT_DOCKER_IMAGE

        # General docker outputs path.
        # According to documentation, all of the python code and python dependencies shall live in "python" dir:
        # https://docs.aws.amazon.com/lambda/latest/dg/configuration-layers.html
        self.outputs_path = '/asset/python'

    def build(self) -> Code:
        # Before building, ensure source code is available for Dockerfile.
        self.__fresh_source_copy()

        # Now, after the source code was made available. Build the code with Docker.
        return Code.from_docker_build(
            path=root,
            build_args={
                # Custom docker image.
                'DOCKER_IMAGE': self.docker_image,

                # OS-level paths.
                'INPUTS_PATH': f'./tmp/{self.source_path_dir_name}',

                # Docker container-level paths.
                'OUTPUTS_PATH': self.outputs_path,

                # Prebuilt commands to install.
                'PIP_INSTALL': self.__dependencies_install_command(),
            },
        )

    def __dependencies_install_command(self) -> str:
        return PipInstall(
            dependencies=self.dependencies,
            additional_pip_install_args=self.additional_pip_install_args,
            output_directory=self.outputs_path
        ).build_command()

    def __fresh_source_copy(self) -> None:
        """
        Copies given lambda layer's source code to a directory where the Dockerfile is.
        This way a Dockerfile can access source code and build it.

        :return: No return.
        """
        # Give a unique directory for every source.
        docker_layer_build_dir = f'{docker_build_root}/{self.source_path_dir_name}'

        # Delete previous copied source.
        shutil.rmtree(docker_layer_build_dir, ignore_errors=True)

        # Copy a fresh source.
        shutil.copytree(
            src=self.source_path,
            # Duplicate parent dir so the source code could be imported as
            # "from parent.module import Module" instead of
            # "from module import Module".
            dst=f'{docker_layer_build_dir}/{self.source_path_dir_name}',
            dirs_exist_ok=True
        )



from aws_cdk.aws_lambda import Function, Code, Runtime
from aws_cdk.core import Stack
from b_aws_testing_framework.tools.cdk_testing.testing_stack import TestingStack

from b_cfn_lambda_layer.lambda_layer import LambdaLayer
from b_cfn_lambda_layer_test.integration.infrastructure.layer1 import root as root1
from b_cfn_lambda_layer_test.integration.infrastructure.layer2 import root as root2
from b_cfn_lambda_layer_test.integration.infrastructure.layer3 import root as root3


class Function3(Function):
    """
    Function that allows us to test whether multiple same layers override each other.
    """
    def __init__(self, scope: Stack):
        super().__init__(
            scope=scope,
            id=f'{TestingStack.global_prefix()}TestingFunction3',
            code=Code.from_inline(
                # Ensure that dummy module is accessible from lambda layer
                # and the parent directory is included.
                'from layer1.dummy_module import DummyModule as M1\n'
                'from layer2.dummy_module import DummyModule as M2\n'
                'from layer3.dummy_module import DummyModule as M3\n'
                '\n\n'
                'def handler(*args, **kwargs):\n'
                '    return dict(\n'
                # Use the dummy module to ensure everything works.
                '        Dummy1=M1.action(),\n'
                '        Dummy2=M2.action(),\n'
                '        Dummy3=M3.action()\n'
                '    )'
                '\n'
            ),
            handler='index.handler',
            runtime=Runtime.PYTHON_3_6,
            layers=[
                LambdaLayer(
                    scope=scope,
                    layer_name=f'{TestingStack.global_prefix()}TestingLayer3_1',
                    source_path=root1,
                    code_runtimes=[Runtime.PYTHON_3_6, Runtime.PYTHON_3_7, Runtime.PYTHON_3_8],
                ),
                LambdaLayer(
                    scope=scope,
                    layer_name=f'{TestingStack.global_prefix()}TestingLayer3_2',
                    source_path=root2,
                    code_runtimes=[Runtime.PYTHON_3_6, Runtime.PYTHON_3_7, Runtime.PYTHON_3_8],
                ),

                # Repeat same layer twice.
                LambdaLayer(
                    scope=scope,
                    layer_name=f'{TestingStack.global_prefix()}TestingLayer3_3_1',
                    source_path=root3,
                    code_runtimes=[Runtime.PYTHON_3_6, Runtime.PYTHON_3_7, Runtime.PYTHON_3_8],
                ),
                LambdaLayer(
                    scope=scope,
                    layer_name=f'{TestingStack.global_prefix()}TestingLayer3_3_2',
                    source_path=root3,
                    code_runtimes=[Runtime.PYTHON_3_6, Runtime.PYTHON_3_7, Runtime.PYTHON_3_8],
                )
            ]
        )



from typing import Optional, List

from b_cfn_lambda_layer.dependency import Dependency


class PipInstall:
    def __init__(
            self,
            dependencies: Optional[List[Dependency]] = None,
            additional_pip_install_args: Optional[str] = None,
            output_directory: Optional[str] = None
    ) -> None:
        """
        Constructor.

        :param dependencies: A dictionary of dependencies to be included within "pip install" command.
            Warning! The dependency strings are prone to code injections!
        :param additional_pip_install_args: A string of additional "pip install" command flags and arguments.
            Warning! This string is prone to code injections!
        """
        self.__dependencies = dependencies or []
        self.__additional_pip_install_args = additional_pip_install_args
        self.__output_directory = output_directory

    def build_command(self) -> str:
        command = ' '.join([dep.build_string() for dep in self.__dependencies if dep.build_string()])

        if command and self.__output_directory:
            command = f'{command} -t {self.__output_directory}'

        if command and self.__additional_pip_install_args:
            command = f'{command} {self.__additional_pip_install_args}'

        if command:
            # If dependencies are given, the "pip install" command will be built.
            command = f'pip install {command}'
        else:
            # If no dependencies are given, a simple "pip list" command is built.
            command = 'pip list'

        return command



from aws_cdk.aws_lambda import Function, Code, Runtime
from aws_cdk.core import Stack
from b_aws_testing_framework.tools.cdk_testing.testing_stack import TestingStack

from b_cfn_lambda_layer.lambda_layer import LambdaLayer
from b_cfn_lambda_layer_test.integration.infrastructure.layer_source import root


class Function2(Function):
    """
    Function that allows us to test whether layer source code is included in the parent directory
    i.e. instead of "from a import A" you would get "from parent_dir.a import A".
    """
    def __init__(self, scope: Stack):
        super().__init__(
            scope=scope,
            id=f'{TestingStack.global_prefix()}TestingFunction2',
            code=Code.from_inline(
                # Ensure that dummy module is accessible from lambda layer
                # and the parent directory is included.
                'from layer_source.dummy_module import DummyModule\n'
                '\n\n'
                'def handler(*args, **kwargs):\n'
                '    return dict(\n'
                # Use the dummy module to ensure everything works.
                '        Dummy=DummyModule.action()\n'
                '    )'
                '\n'
            ),
            handler='index.handler',
            runtime=Runtime.PYTHON_3_6,
            layers=[
                LambdaLayer(
                    scope=scope,
                    layer_name=f'{TestingStack.global_prefix()}TestingLayer2',
                    source_path=root,
                    code_runtimes=[Runtime.PYTHON_3_6, Runtime.PYTHON_3_7, Runtime.PYTHON_3_8],
                )
            ]
        )



from aws_cdk.aws_lambda import Function, Code, Runtime
from aws_cdk.core import Stack
from b_aws_testing_framework.tools.cdk_testing.testing_stack import TestingStack

from b_cfn_lambda_layer.lambda_layer import LambdaLayer
from b_cfn_lambda_layer.package_version import PackageVersion
from b_cfn_lambda_layer_test.integration.infrastructure.layer_source import root


class Function1(Function):
    """
    Function that allows us to test whether installing dependencies works.
    """
    def __init__(self, scope: Stack):
        super().__init__(
            scope=scope,
            id=f'{TestingStack.global_prefix()}TestingFunction1',
            code=Code.from_inline(
                'import urllib3\n'
                'import jose\n'
                'from jose import jwk, jwt\n'
                'from jose.utils import base64url_decode\n'
                'import boto3\n'
                'import botocore\n'
                'from layer_source.dummy_module import DummyModule\n'
                '\n\n'
                'def handler(*args, **kwargs):\n'
                '    return dict(\n'
                '        Boto3Version=boto3.__version__,\n'
                '        BotocoreVersion=botocore.__version__,\n'
                '        JoseVersion=jose.__version__,\n'
                '        Dummy=DummyModule.action()\n'
                '    )'
                '\n'
            ),
            handler='index.handler',
            runtime=Runtime.PYTHON_3_6,
            layers=[
                LambdaLayer(
                    scope=scope,
                    layer_name=f'{TestingStack.global_prefix()}TestingLayer1',
                    source_path=root,
                    code_runtimes=[Runtime.PYTHON_3_6, Runtime.PYTHON_3_7, Runtime.PYTHON_3_8],
                    dependencies={
                        'python-jose': PackageVersion.from_string_version('3.3.0'),
                        'boto3': PackageVersion.from_string_version('1.16.35'),
                        'botocore': PackageVersion.from_string_version('1.19.35')
                    },
                )
            ]
        )



from aws_cdk.core import Construct
from b_aws_testing_framework.tools.cdk_testing.testing_stack import TestingStack

from b_cfn_lambda_layer_test.integration.infrastructure.function1 import Function1
from b_cfn_lambda_layer_test.integration.infrastructure.function2 import Function2
from b_cfn_lambda_layer_test.integration.infrastructure.function3 import Function3


class MainStack(TestingStack):
    LAMBDA_FUNCTION_1_NAME_KEY = 'LambdaFunction1Name'
    LAMBDA_FUNCTION_2_NAME_KEY = 'LambdaFunction2Name'
    LAMBDA_FUNCTION_3_NAME_KEY = 'LambdaFunction3Name'

    def __init__(self, scope: Construct):
        super().__init__(scope=scope)

        self.function1 = Function1(self)
        self.function2 = Function2(self)
        self.function3 = Function3(self)

        self.add_output(self.LAMBDA_FUNCTION_1_NAME_KEY, value=self.function1.function_name)
        self.add_output(self.LAMBDA_FUNCTION_2_NAME_KEY, value=self.function2.function_name)
        self.add_output(self.LAMBDA_FUNCTION_3_NAME_KEY, value=self.function3.function_name)



import logging
from typing import List, Optional, Dict

from aws_cdk.aws_lambda import LayerVersion, Runtime
from aws_cdk.core import DockerImage
from aws_cdk.core import Stack

from b_cfn_lambda_layer.dependency import Dependency
from b_cfn_lambda_layer.lambda_layer_code import LambdaLayerCode
from b_cfn_lambda_layer.package_version import PackageVersion

LOGGER = logging.getLogger(__name__)


class LambdaLayer(LayerVersion):
    def __init__(
            self,
            scope: Stack,
            layer_name: str,
            source_path: str,
            code_runtimes: List[Runtime],
            dependencies: Optional[Dict[str, PackageVersion]] = None,
            additional_pip_install_args: Optional[str] = None,
            docker_image: Optional[DockerImage] = None,
    ) -> None:
        """
        Constructor.

        :param scope: Parent CloudFormation stack.
        :param layer_name: Unique name of the layer.
        :param source_path: Path to source-code to be bundled.
        :param code_runtimes: Available runtimes for your code.
        :param dependencies: A dictionary of dependencies to include in the layer.
            Keys are dependency (package) names.
            Values are dependency (package) version objects.
        :param additional_pip_install_args: A string of additional pip-install arguments.
        :param docker_image: Docker image to use when building code.
        """
        super().__init__(
            scope=scope,
            id=layer_name,
            layer_version_name=layer_name,
            code=LambdaLayerCode(
                source_path=source_path,
                additional_pip_install_args=additional_pip_install_args,
                dependencies=[Dependency(key, value) for key, value in (dependencies or {}).items()],
                docker_image=docker_image
            ).build(),
            compatible_runtimes=code_runtimes
        )
